@@ -17,10 +17,17 @@
  * with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#define G_LOG_DOMAIN "GcalEventWidget"
+
+#include <string.h>
+#include <glib/gi18n.h>
+
 #include "gcal-event-widget.h"
 #include "gcal-utils.h"
 
-#define INTENSITY(c) ((c->red) * 0.30 + (c->green) * 0.59 + (c->blue) * 0.11)
+#define DESC_MAX_CHAR 200
+#define INTENSITY(c)  ((c->red) * 0.30 + (c->green) * 0.59 + (c->blue) * 0.11)
+#define ICON_SIZE     16
 
 struct _GcalEventWidget
 {
@@ -37,8 +44,7 @@ struct _GcalEventWidget
 
   GcalEvent     *event;
 
-  /* Drag n' Drop icon */
-  GdkPixbuf     *dnd_pixbuf;
+  GtkOrientation orientation;
 
   GdkWindow     *event_window;
   gboolean       button_pressed;
@@ -50,6 +56,7 @@ enum
   PROP_DATE_END,
   PROP_DATE_START,
   PROP_EVENT,
+  PROP_ORIENTATION,
   NUM_PROPS
 };
 
@@ -61,7 +68,95 @@ enum
 
 static guint signals[NUM_SIGNALS] = { 0, };
 
-G_DEFINE_TYPE (GcalEventWidget, gcal_event_widget, GTK_TYPE_WIDGET)
+G_DEFINE_TYPE_WITH_CODE (GcalEventWidget, gcal_event_widget, GTK_TYPE_WIDGET,
+                         G_IMPLEMENT_INTERFACE (GTK_TYPE_ORIENTABLE, NULL));
+
+static gchar*
+get_visible_text (GtkWidget         *widget,
+                  GtkOrientation     orientation)
+{
+  GcalEventWidget *self;
+  gboolean is_ltr;
+  gchar *display_text;
+  gchar *escaped_text;
+
+  self = GCAL_EVENT_WIDGET (widget);
+  is_ltr = gtk_widget_get_direction (widget) != GTK_TEXT_DIR_RTL;
+
+  /*
+   * Since we set markup labels, we have to escape text before
+   * using it, or we get empty event widgets.
+   */
+  escaped_text = g_markup_escape_text (gcal_event_get_summary (self->event), -1);
+
+  /*
+   * The text displayed in the widget includes the time when the event
+   * is timed, or not if the event is all day.
+   */
+  if (gcal_event_get_all_day (self->event))
+    {
+      display_text = escaped_text;
+    }
+  else
+    {
+      GDateTime *local_start_time;
+      gchar *start_time;
+      local_start_time = g_date_time_to_local (gcal_event_get_date_start (self->event));
+
+      if (self->clock_format_24h)
+        start_time = g_date_time_format (local_start_time, "%R");
+      else
+        start_time = g_date_time_format (local_start_time, "%I:%M %P");
+
+      /*
+       * When in horizontal, draw the elements sequentially; otherwise,
+       * draw the time in the first line, and the title in subsequent
+       * lines.
+       */
+      if (orientation == GTK_ORIENTATION_HORIZONTAL)
+        {
+          if (is_ltr)
+            display_text = g_strdup_printf ("(%s) %s", start_time, escaped_text);
+          else
+            display_text = g_strdup_printf ("%s (%s)", escaped_text, start_time);
+        }
+      else
+        {
+          display_text = g_strdup_printf ("<b>%s</b>\n%s", start_time, escaped_text);
+        }
+
+      g_clear_pointer (&local_start_time, g_date_time_unref);
+      g_clear_pointer (&escaped_text, g_free);
+      g_clear_pointer (&start_time, g_free);
+    }
+
+    return display_text;
+}
+
+static gint
+get_vertical_text_height (GtkWidget         *widget)
+{
+  GcalEventWidget *self;
+  GtkBorder border, padding;
+  GtkStyleContext *context;
+  PangoLayout *layout;
+  gint layout_height;
+  gchar* display_text;
+
+  self = GCAL_EVENT_WIDGET (widget);
+  context = gtk_widget_get_style_context (widget);
+  display_text = get_visible_text (widget, GTK_ORIENTATION_VERTICAL);
+  layout = gtk_widget_create_pango_layout (widget, display_text);
+  pango_layout_get_pixel_size (layout, NULL, &layout_height);
+  g_object_unref (layout);
+
+  gtk_style_context_get_border (context, gtk_style_context_get_state (context), &border);
+  gtk_style_context_get_padding (context, gtk_style_context_get_state (context), &padding);
+
+  g_clear_pointer(&display_text, g_free);
+
+  return layout_height + padding.top + padding.bottom + border.top + border.bottom;
+}
 
 static void
 gcal_event_widget_update_style (GcalEventWidget *self)
@@ -147,6 +242,157 @@ update_color (GcalEventWidget *self)
   g_free (color_str);
 }
 
+static void
+gcal_event_widget_set_event_tooltip (GcalEventWidget *self,
+                                     GcalEvent       *event)
+{
+  g_autoptr (GDateTime) tooltip_start, tooltip_end;
+  g_autofree gchar *start, *end, *escaped_summary;
+  GString *tooltip_mesg;
+  gboolean allday, multiday, is_ltr;
+  guint description_len;
+
+  tooltip_mesg = g_string_new (NULL);
+  escaped_summary = g_markup_escape_text (gcal_event_get_summary (event), -1);
+  g_string_append_printf (tooltip_mesg, "<b>%s</b>", escaped_summary);
+
+  tooltip_start = g_date_time_to_local (gcal_event_get_date_start (event));
+  tooltip_end = g_date_time_to_local (gcal_event_get_date_end (event));
+
+  allday = gcal_event_get_all_day (event);
+  multiday = gcal_event_is_multiday (event);
+
+  is_ltr = gtk_widget_get_direction (GTK_WIDGET (self)) != GTK_TEXT_DIR_RTL;
+
+  if (allday)
+    {
+      if (multiday)
+        {
+          start = g_date_time_format (tooltip_start, "%x");
+          end = g_date_time_format (tooltip_end, "%x");
+        }
+      else
+        {
+          start = g_date_time_format (tooltip_start, "%x");
+          end = NULL;
+        }
+    }
+  else
+    {
+      if (multiday)
+        {
+          if (self->clock_format_24h)
+            {
+              if (is_ltr)
+                {
+                  start = g_date_time_format (tooltip_start, "%x %R");
+                  end = g_date_time_format (tooltip_end, "%x %R");
+                }
+              else
+                {
+                  start = g_date_time_format (tooltip_start, "%R %x");
+                  end = g_date_time_format (tooltip_end, "%R %x");
+                }
+            }
+          else
+            {
+              if (is_ltr)
+                {
+                  start = g_date_time_format (tooltip_start, "%x I:%M %P");
+                  end = g_date_time_format (tooltip_end, "%x I:%M %P");
+                }
+              else
+                {
+                  start = g_date_time_format (tooltip_start, "%P %M:%I %x");
+                  end = g_date_time_format (tooltip_end, "%P %M:%I %x");
+                }
+            }
+        }
+      else
+        {
+          if (self->clock_format_24h)
+            {
+              if (is_ltr)
+                {
+                  start = g_date_time_format (tooltip_start, "%x, %R");
+                  end = g_date_time_format (tooltip_end, "%R");
+                }
+              else
+                {
+                  start = g_date_time_format (tooltip_start, "%R ,%x");
+                  end = g_date_time_format (tooltip_end, "%R");
+                }
+            }
+          else
+            {
+              if (is_ltr)
+                {
+                  start = g_date_time_format (tooltip_start, "%x, I:%M %P");
+                  end = g_date_time_format (tooltip_end, "I:%M %P");
+                }
+              else
+                {
+                  start = g_date_time_format (tooltip_start, "%P %M:%I ,%x");
+                  end = g_date_time_format (tooltip_end, "%P %M:%I");
+                }
+            }
+        }
+    }
+
+  if (allday && !multiday)
+    {
+      g_string_append_printf (tooltip_mesg,
+                              "\n%s",
+                              start);
+    }
+  else
+    {
+      g_string_append_printf (tooltip_mesg,
+                              "\n%s - %s",
+                              is_ltr ? start : end,
+                              is_ltr ? end : start);
+    }
+
+  /* Append event location */
+  if (g_utf8_strlen (gcal_event_get_location (event), -1) > 0)
+    {
+      g_autofree gchar *escaped_location;
+
+      escaped_location = g_markup_escape_text (gcal_event_get_location (event), -1);
+
+      g_string_append (tooltip_mesg, "\n\n");
+      g_string_append_printf (tooltip_mesg, "%s", escaped_location);
+    }
+
+  description_len = g_utf8_strlen (gcal_event_get_description (event), -1);
+
+  /* Truncate long descriptions at a white space and ellipsize */
+  if (description_len > 0)
+    {
+      g_autofree gchar *escaped_description;
+      GString *tooltip_desc;
+
+      tooltip_desc = g_string_new (gcal_event_get_description (event));
+
+      /* If the description is larger than DESC_MAX_CHAR, ellipsize it */
+      if (description_len > DESC_MAX_CHAR)
+        {
+          g_string_truncate (tooltip_desc, DESC_MAX_CHAR - 1);
+          g_string_append (tooltip_desc, "…");
+        }
+
+      escaped_description = g_markup_escape_text (tooltip_desc->str, -1);
+
+      g_string_append_printf (tooltip_mesg, "\n\n%s", escaped_description);
+
+      g_string_free (tooltip_desc, TRUE);
+    }
+
+  gtk_widget_set_tooltip_markup (GTK_WIDGET (self), tooltip_mesg->str);
+
+  g_string_free (tooltip_mesg, TRUE);
+}
+
 static void
 gcal_event_widget_set_event_internal (GcalEventWidget *self,
                                       GcalEvent       *event)
@@ -182,19 +428,18 @@ gcal_event_widget_set_event_internal (GcalEventWidget *self,
                             "notify::summary",
                             G_CALLBACK (gtk_widget_queue_draw),
                             self);
+
+  gcal_event_widget_set_event_tooltip (self, event);
 }
 
 static void
 gcal_event_widget_init (GcalEventWidget *self)
 {
   GtkWidget *widget;
-  GSettings *settings;
-  gchar *clock_format;
 
   widget = GTK_WIDGET (self);
-  settings = g_settings_new ("org.gnome.desktop.interface");
-  clock_format = g_settings_get_string (settings, "clock-format");
-  self->clock_format_24h = g_strcmp0 (clock_format, "24h") == 0;
+  self->clock_format_24h = is_clock_format_24h ();
+  self->orientation = GTK_ORIENTATION_HORIZONTAL;
 
   gtk_widget_set_has_window (widget, FALSE);
   gtk_widget_set_can_focus (widget, TRUE);
@@ -204,12 +449,9 @@ gcal_event_widget_init (GcalEventWidget *self)
                        GDK_BUTTON1_MASK,
                        NULL,
                        0,
-                       GDK_ACTION_COPY);
+                       GDK_ACTION_MOVE);
 
   gtk_drag_source_add_text_targets (widget);
-
-  g_clear_object (&settings);
-  g_free (clock_format);
 }
 
 static void
@@ -234,6 +476,12 @@ gcal_event_widget_set_property (GObject      *object,
       gcal_event_widget_set_event_internal (self, g_value_get_object (value));
       break;
 
+    case PROP_ORIENTATION:
+      self->orientation = g_value_get_enum (value);
+      gtk_widget_queue_draw (GTK_WIDGET (object));
+      g_object_notify (object, "orientation");
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -261,6 +509,10 @@ gcal_event_widget_get_property (GObject      *object,
       g_value_set_object (value, self->event);
       break;
 
+    case PROP_ORIENTATION:
+      g_value_set_enum (value, self->orientation);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -279,7 +531,6 @@ gcal_event_widget_finalize (GObject *object)
 
   /* releasing properties */
   g_clear_pointer (&self->css_class, g_free);
-  g_clear_object (&self->dnd_pixbuf);
   g_clear_object (&self->event);
 
   G_OBJECT_CLASS (gcal_event_widget_parent_class)->finalize (object);
@@ -314,13 +565,17 @@ gcal_event_widget_get_preferred_height (GtkWidget *widget,
                                         gint      *minimum,
                                         gint      *natural)
 {
+  GcalEventWidget *self;
   GtkBorder border, padding;
   GtkStyleContext *context;
   PangoLayout *layout;
   gint layout_height;
+  gchar* display_text;
 
+  self = GCAL_EVENT_WIDGET (widget);
   context = gtk_widget_get_style_context (widget);
-  layout = gtk_widget_create_pango_layout (widget, NULL);
+  display_text = get_visible_text (widget, self->orientation);
+  layout = gtk_widget_create_pango_layout (widget, display_text);
   pango_layout_get_pixel_size (layout, NULL, &layout_height);
   g_object_unref (layout);
 
@@ -331,6 +586,8 @@ gcal_event_widget_get_preferred_height (GtkWidget *widget,
     *minimum = layout_height + padding.top + padding.bottom + border.top + border.bottom;
   if (natural != NULL)
     *natural = layout_height + padding.top + padding.bottom + border.top + border.bottom;
+
+  g_clear_pointer(&display_text, g_free);
 }
 
 static void
@@ -475,46 +732,26 @@ gcal_event_widget_draw (GtkWidget *widget,
   gtk_render_background (context, cr, 0, 0, width, height);
   gtk_render_frame (context, cr, 0, 0, width, height);
 
-  /*
-   * The text displayed in the widget includes the time when the event
-   * is timed, or not if the event is all day.
-   */
-  if (gcal_event_get_all_day (self->event))
-    {
-      display_text = g_strdup (gcal_event_get_summary (self->event));
-    }
+  /* Set the text to be displayed */
+  if (self->orientation == GTK_ORIENTATION_HORIZONTAL)
+    display_text = get_visible_text (widget, GTK_ORIENTATION_HORIZONTAL);
   else
     {
-      GDateTime *local_start_time;
-      gchar *start_time;
-      local_start_time = g_date_time_to_local (gcal_event_get_date_start (self->event));
-
-      if (self->clock_format_24h)
-        start_time = g_date_time_format (local_start_time, "%R");
+      if (get_vertical_text_height (widget) >= height)
+        display_text = get_visible_text (widget, GTK_ORIENTATION_HORIZONTAL);
       else
-        start_time = g_date_time_format (local_start_time, "%I:%M %P");
-
-      if (is_ltr)
-        display_text = g_strdup_printf ("(%s) %s", start_time, gcal_event_get_summary (self->event));
-      else
-        display_text = g_strdup_printf ("%s (%s)", gcal_event_get_summary (self->event), start_time);
-
-      g_clear_pointer (&local_start_time, g_date_time_unref);
-      g_clear_pointer (&start_time, g_free);
+        display_text = get_visible_text (widget, GTK_ORIENTATION_VERTICAL);
     }
 
   /* Retrieve the font description */
   gtk_style_context_get (context, state, "font", &font_desc, NULL);
-  layout = gtk_widget_create_pango_layout (widget, display_text);
+  layout = gtk_widget_create_pango_layout (widget, "");
+  pango_layout_set_markup (layout, display_text, -1);
   pango_layout_set_font_description (layout, font_desc);
   pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
-  pango_layout_set_width (layout, (width - (padding.left + padding.right) ) * PANGO_SCALE);
-  pango_cairo_update_layout (cr, layout);
 
   /* Icon size */
-  pango_layout_get_pixel_size (layout, NULL, &layout_height);
-  icon_size = 4 * (layout_height / 4);
-
+  icon_size = ICON_SIZE;
   end_gap = 0;
 
   if (gcal_event_has_alarms (self->event))
@@ -526,8 +763,36 @@ gcal_event_widget_draw (GtkWidget *widget,
   x = is_ltr ? padding.left : padding.left + end_gap;
 
   pango_layout_set_width (layout, (width - (padding.left + padding.right + end_gap)) * PANGO_SCALE);
+  pango_layout_get_pixel_size (layout, NULL, &layout_height);
   gtk_render_layout (context, cr, x, padding.top, layout);
 
+#if 0
+  /* TODO: find out a good way to handle this */
+  /*
+   * On vertical event widgets, we assume we might have space to draw the
+   * event's detail (or at least part of it).
+   */
+  if (self->orientation == GTK_ORIENTATION_VERTICAL &&
+      2 * padding.top + layout_height + padding.bottom < height &&
+      strlen (gcal_event_get_description (self->event)) > 0)
+    {
+      g_autofree gchar *description;
+
+      description = g_strdup_printf ("<small>%s</small>", gcal_event_get_description (self->event));
+      pango_layout_set_markup (layout, description, -1);
+      pango_layout_set_width (layout, (width - (padding.left + padding.right) ) * PANGO_SCALE);
+      pango_layout_set_height (layout, (height - (2 * padding.top + layout_height + padding.bottom) ) * PANGO_SCALE);
+      pango_layout_set_font_description (layout, font_desc);
+      pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
+
+      gtk_render_layout (context,
+                         cr,
+                         x,
+                         2 * padding.top + layout_height + padding.bottom,
+                         layout);
+    }
+#endif
+
   x += is_ltr ? (width - end_gap - padding.left - padding.right) : (- padding.right - padding.left - icon_size);
 
   /* render reminder icon */
@@ -548,7 +813,11 @@ gcal_event_widget_draw (GtkWidget *widget,
                                                         &was_symbolic,
                                                         NULL);
 
-      gdk_cairo_set_source_pixbuf (cr, pixbuf, x + padding.left, padding.top);
+      gtk_render_icon (context,
+                       cr,
+                       pixbuf,
+                       x + padding.left,
+                       padding.top);
 
       x += is_ltr ? (padding.left + icon_size) : (- padding.right - icon_size);
 
@@ -574,7 +843,12 @@ gcal_event_widget_draw (GtkWidget *widget,
                                                         &was_symbolic,
                                                         NULL);
 
-      gdk_cairo_set_source_pixbuf (cr, pixbuf, x + padding.left, padding.top);
+      gtk_render_icon (context,
+                       cr,
+                       pixbuf,
+                       x + padding.left,
+                       padding.top);
+
       g_object_unref (pixbuf);
       cairo_paint (cr);
     }
@@ -594,7 +868,6 @@ gcal_event_widget_button_press_event (GtkWidget      *widget,
 
   self = GCAL_EVENT_WIDGET (widget);
   self->button_pressed = TRUE;
-
   return TRUE;
 }
 
@@ -616,13 +889,31 @@ gcal_event_widget_button_release_event (GtkWidget      *widget,
   return FALSE;
 }
 
+static cairo_surface_t*
+get_dnd_icon (GtkWidget *widget)
+{
+  cairo_surface_t *surface;
+  cairo_t *cr;
+
+  /* Make it transparent */
+  surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
+                                        gtk_widget_get_allocated_width (widget),
+                                        gtk_widget_get_allocated_height (widget));
+
+  cr = cairo_create (surface);
+
+  gtk_widget_draw (widget, cr);
+  cairo_destroy (cr);
+
+  return surface;
+}
+
 static void
 gcal_event_widget_drag_begin (GtkWidget      *widget,
                               GdkDragContext *context)
 {
   GcalEventWidget *self = GCAL_EVENT_WIDGET (widget);
-
-  g_clear_object (&self->dnd_pixbuf);
+  cairo_surface_t *surface;
 
   if (self->read_only)
     {
@@ -631,13 +922,11 @@ gcal_event_widget_drag_begin (GtkWidget      *widget,
     }
 
   /* Setup the drag n' drop icon */
-  self->dnd_pixbuf = gdk_pixbuf_get_from_window (self->event_window,
-                                                 0,
-                                                 0,
-                                                 gtk_widget_get_allocated_width (widget),
-                                                 gtk_widget_get_allocated_height (widget));
+  surface = get_dnd_icon (widget);
 
-  gtk_drag_set_icon_pixbuf (context, self->dnd_pixbuf, 0, 0);
+  gtk_drag_set_icon_surface (context, surface);
+
+  g_clear_pointer (&surface, cairo_surface_destroy);
 }
 
 static gboolean
@@ -717,6 +1006,13 @@ gcal_event_widget_class_init(GcalEventWidgetClass *klass)
                                                         GCAL_TYPE_EVENT,
                                                         G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 
+  /**
+   * GcalEventWidget::orientation:
+   *
+   * The orientation of the event widget.
+   */
+  g_object_class_override_property (object_class, PROP_ORIENTATION, "orientation");
+
   signals[ACTIVATE] = g_signal_new ("activate",
                                      GCAL_TYPE_EVENT_WIDGET,
                                      G_SIGNAL_RUN_LAST,
@@ -753,7 +1049,7 @@ gcal_event_widget_set_read_only (GcalEventWidget *event,
                            0,
                            NULL,
                            0,
-                           GDK_ACTION_COPY);
+                           GDK_ACTION_MOVE);
 
       gtk_drag_source_add_text_targets (widget);
     }
